#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "BasicScheme.h"
#include "RandomGenerator.h"
#include "SampleG.h"
#include "SampleP.h"
#include "Hash.h"
#include "fft.h"

long long n, k, modulo;
double r, s;

extern poly_R hash_value;

extern poly_R _hash_value;

void parameter_init(int lambda) {
	extern long long modulo; 
	
	if(lambda == 128) {
		modulo = 16770049;
		n = 512;
		k = 24;
		r = 5.6;
		s = 44873;
		
		FFT_init(9, 88150, 5793134, -32754);
	} else if(lambda == 256) {
		modulo = 268369921;
		n = 1024;
		k = 28;
		r = 7.8;
		s = 131360;
		
		FFT_init(10, 326097, 228572460, -262080);
	}
}

void KeyGen(mat_poly_R **_a, mat_poly_R **_e, mat_poly_R **_r) {
	*_e = (mat_poly_R*)malloc(sizeof(mat_poly_R));
	**_e = generate_Gaussian_mat_poly_R(n, 1, k, 2 * sqrt(n));
	
	*_r = (mat_poly_R*)malloc(sizeof(mat_poly_R));
	**_r = generate_Gaussian_mat_poly_R(n, 1, k, 2 * sqrt(n));
	
	*_a = (mat_poly_R*)malloc(sizeof(mat_poly_R));
	
	**_a = (mat_poly_R){1, k + 2, NULL};
	(*_a) -> coefficient = (poly_R*)malloc(sizeof(poly_R) * 1 * (k + 2));

	*((*_a) -> coefficient) = create_poly_R(n);
	*((*_a) -> coefficient -> coefficient) = 1;
	
	*((*_a) -> coefficient + 1) = generate_random_poly_R(n);
	
	long long i;
	for(i = 0; i < k; i++) {
		*((*_a) -> coefficient + 2 + i) = create_poly_R(n);
		*(((*_a) -> coefficient + 2 + i) -> coefficient) = 1 << i;
		
		*((*_a) -> coefficient + 2 + i) = mod_sub_poly_R(*((*_a) -> coefficient + 2 + i), mod_add_poly_R
			(mod_mul_poly_R(*((*_a) -> coefficient + 1), *((*_r) -> coefficient + i)), *((*_e) -> coefficient + i)));
	}
}

mat_poly_R Sign(const mat_poly_R *_a, const mat_poly_R *_e, const mat_poly_R *_r, const unsigned char* message, long long message_length) {
	extern long long modulo;

	//mat_poly_R _p = generate_Gaussian_mat_poly_R(n, 1, k + 2, 2 * sqrt(n));
	//mat_poly_R _p = create_mat_poly_R(n, 1, k + 2);
	mat_poly_R _p = Sample_P(n, modulo, s, r, *_e, *_r);
	
	//hash_value = generate_random_poly_R(n);
	//hash_value = create_poly_R(n);
	poly_R hash_value = Hash(message, message_length, n);
	
	poly_R v = mod_sub_poly_R(hash_value, inner_product_mat_poly_R(*_a, _p));
	
	mat_poly_R _x = create_mat_poly_R(n, 1, k);
	long long *z = (long long*)malloc(sizeof(long long) * k);
	
	long long i;
	for(i = 0; i < n; i++) {
		Sample_G(z, modulo, *(v.coefficient + i), r);
		
		long long j;
		for(j = 0; j < k; j++) {
			*((_x.coefficient + j) -> coefficient + i) = *(z + j);
		}
	}
	
	*(_p.coefficient) = mod_add_poly_R(*(_p.coefficient), inner_product_mat_poly_R(*_e, _x));
	*(_p.coefficient + 1) = mod_add_poly_R(*(_p.coefficient + 1), inner_product_mat_poly_R(*_r, _x));
	for(i = 0; i < k; i++) {
		*(_p.coefficient + 2 + i) = mod_add_poly_R(*(_p.coefficient + 2 + i), *(_x.coefficient + i));
	}
	
	return _p;
}

int Verify(const mat_poly_R *_a, const mat_poly_R *sig, const unsigned char *message, long long message_length) {
	poly_R _hash_value = inner_product_mat_poly_R(*_a, *sig);
	poly_R hash_value = Hash(message, message_length, n);
	
	long long i;
	for(i = 0; i < hash_value.degree; i++) {
		if(*(_hash_value.coefficient + i) != *(hash_value.coefficient + i)) {
			return 0;
		}
	}
	
	return 1;
}
