#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "RandomGenerator.h"

#define PI 3.1415926

extern long long modulo;

long long generate_randomBnd(long long q) {
	return (((rand() << 16) + rand()) % q) - q / 2;
}

double generate_Gaussian(double mean, double standard_deviation) {
	//Box-Muller
	double u_1, u_2, theta, r;

	u_1 = (double)rand() / RAND_MAX;
	u_2 = (double)rand() / RAND_MAX;

	theta = 2 * PI * u_1;
	r = sqrt(-2 * log(u_2));

	return mean + standard_deviation * r * cos(theta);
}

long long Oracle_1(double c, double standard_deviation) {
	double s = generate_Gaussian(c, standard_deviation);

	return s >= 0 ? s + 0.5 : s - 0.5;
}

long long Oracle_2(long long coset, double standard_deviation) {
	double s = generate_Gaussian(0, standard_deviation);

	long long result = s;

	if(coset % 2 == 0) {
		if(result % 2 != 0) {
			result = result > 0 ? result + 1 : result - 1;
		}
	} else {
		if(result % 2 == 0) {
			if(result == 0) {
				result = s > 0 ? 1 : -1;
			} else {
				result = result > 0 ? result + 1 : result - 1;
			}
		}
	}

	return result;
}

poly_R generate_random_poly_R(long long n) {
	poly_R result = (poly_R){n, NULL};
	result.coefficient = (long long*)malloc(sizeof(long long) * n);
	
	long long i;
	for(i = 0; i < n; i++) {
		*(result.coefficient + i) = generate_randomBnd(modulo);
	}
	
	return result;
}

mat_poly_R generate_Gaussian_mat_poly_R(long long n, long long a, long long b, double sigma) {
	mat_poly_R result = (mat_poly_R){a, b, NULL};
	result.coefficient = (poly_R*)malloc(sizeof(poly_R) * a * b);
	
	long long i;
	for(i = 0; i < a; i++) {
		long long j;
		for(j = 0; j < b; j++) {
			*(result.coefficient + i * b + j) = (poly_R){n, NULL};
			(result.coefficient + i * b + j) -> coefficient = (long long*)malloc(sizeof(long long) * n);
			
			long long k;
			for(k = 0; k < n; k++) {
				*((result.coefficient + i * b + j) -> coefficient + k) = Oracle_1(0, sigma);
			}
		}
	}
	
	return result;
}

poly_K generate_Gaussian_poly_K(long long n, double sigma) {
	poly_K result = create_poly_K(n);
	
	long long i;
	for(i = 0; i < result.degree; i++) {
		*(result.coefficient + i) = generate_Gaussian(0, sigma);
	}
	
	return result;
}
