#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fft.h"

int *rev;
long long *roots;
long long *inv_roots;
long long _n;

extern long long modulo;

void FFT_init(int lambda, long long root, long long inv_root, long long inv_n) {
	rev = (int*)malloc(sizeof(int) * (1 << lambda));
	rev[0] = 0;
	
	roots = (long long*)malloc(sizeof(long long) * lambda);
	inv_roots = (long long*)malloc(sizeof(long long) * lambda);
	roots[lambda - 1] = root;
	inv_roots[lambda - 1] = inv_root;
	
	int i;
	for(i = 0; i < (1 << lambda); i++) {
		rev[i] = (rev[i >> 1] >> 1) | ((i & 1) << (lambda - 1));
	}
	
	for(i = 1; i < lambda; i++) {
		roots[lambda - 1 - i] = (roots[lambda - i] * roots[lambda - i]) % modulo;
		inv_roots[lambda - 1 - i] = (inv_roots[lambda - i] * inv_roots[lambda - i]) % modulo;
	}
	
	_n = inv_n;
}

void print_Complex(Complex a) {
	printf("%f + %f i", a.a, a.b);
}

Complex add_Complex_Complex(Complex a, Complex b) {
	return (Complex){a.a + b.a, a.b + b.b};
}

Complex sub_Complex_Complex(Complex a, Complex b) {
	return (Complex){a.a - b.a, a.b - b.b};
}

Complex mul_Complex_Complex(Complex a, Complex b) {
	return (Complex){a.a * b.a - a.b * b.b, a.a * b.b + a.b * b.a};
}

Complex div_Complex_Complex(Complex a, Complex b) {
	if(b.a == 0 && b.b == 0) {
		return (Complex){0, 0};
	} else {
		return (Complex){(a.a * b.a + a.b * b.b) / (b.a * b.a + b.b * b.b), (a.b * b.a - a.a * b.b) / (b.a * b.a + b.b * b.b)};
	}
}

/*void FFT_R(long long *z, Complex *c, int len, int opt) {
	Complex *a = (Complex*)malloc(sizeof(Complex) * len);
	
	int i;
	for(i = 0; i < len; i++) {
		if(opt == 1) {
			a[i] = (Complex){z[rev[i]], 0};
		} else if(opt == -1) {
			a[i] = c[rev[i]];
		}
	}
	
	for(i = 1; i < len; i <<= 1) {
		Complex wn = (Complex){cos(PI / i), opt * sin(PI / i)};
		
		int j;
		for(j = 0; j < len; j += (i << 1)) {
			Complex w;
			
			if(opt == 1) {
				w = (Complex){cos(PI / (2 * i)), sin(PI / (2 * i))};
			} else if(opt == -1) {
				w = (Complex){1, 0};
			}
			
			int k;
			for(k = 0; k < i; k++, w = mul_Complex_Complex(w, wn)) {
				Complex x = a[j + k];
				Complex y = mul_Complex_Complex(w, a[j + k + i]);
				a[j + k] = add_Complex_Complex(x, y);
				a[j + k + i] = sub_Complex_Complex(x, y);
			}
		}
	}
	
	for(i = 0; i < len; i++) {
		if(opt == 1) {
			c[i] = a[i];
		} else if(opt == -1) {
			Complex t = mul_Complex_Complex(a[i], (Complex){cos(i * PI / len), -sin(i * PI / len)});
			z[i] = (t.a > 0 ? t.a + 0.5 : t.a - 0.5) / len;
		}
	}
}*/

void FFT_R(long long *input, long long *output, int len, int opt) {
	long long *a = (long long*)malloc(sizeof(long long) * len);
	
	int i;
	for(i = 0; i < len; i++) {
		a[i] = input[rev[i]];
	}
	
	int d;
	for(i = 1, d = 0; i < len; i <<= 1, d++) {
		long long gn;
		if(opt == 1) {
			gn = (roots[d] * roots[d]) % modulo;
		} else if(opt == -1) {
			gn = (inv_roots[d] * inv_roots[d]) % modulo;
		}
		
		int j;
		for(j = 0; j < len; j += (i << 1)) {
			long long g;
			if(opt == 1) {
				g = roots[d];
			} else if(opt == -1) {
				g = 1;
			}
			
			int k;
			for(k = 0; k < i; k++, g = (g * gn) % modulo) {
				long long x = a[j + k];
				long long y = (g * a[j + k + i]) % modulo;
				a[j + k] = (x + y) % modulo;
				a[j + k + i] = (x - y) % modulo;
			}
		}
	}
	
	long long r = 1;
	
	for(i = 0; i < len; i++) {
		if(opt == 1) {
			output[i] = a[i] >= 0 ? (a[i] % modulo > modulo / 2 ? a[i] % modulo - modulo : a[i] % modulo) : (a[i] % modulo < -modulo / 2 ? a[i] % modulo + modulo : a[i] % modulo);
		} else if(opt == -1) {
			output[i] = (a[i] * r) % modulo;
			output[i] = (output[i] * _n) % modulo;
			output[i] = output[i] >= 0 ? (output[i] > modulo / 2 ? output[i] - modulo : output[i]) : (output[i] < -modulo / 2 ? output[i] + modulo : output[i]);
			r = (r * inv_roots[d - 1]) % modulo;
		}
	}
}

void FFT_K(double *z, Complex *c, int len, int opt) {
	Complex *a = (Complex*)malloc(sizeof(Complex) * len);

	int i;
	for(i = 0; i < len; i++) {
		if(opt == 1) {
			a[i] = (Complex){z[rev[i]], 0};
		} else if(opt == -1) {
			a[i] = c[rev[i]];
		}
	}

	for(i = 1; i < len; i <<= 1) {
		Complex wn = (Complex){cos(PI / i), opt * sin(PI / i)};

		int j;
		for(j = 0; j < len; j += (i << 1)) {
			Complex w;

			if(opt == 1) {
				w = (Complex){cos(PI / (2 * i)), sin(PI / (2 * i))};
			} else if(opt == -1) {
				w = (Complex){1, 0};
			}

			int k;
			for(k = 0; k < i; k++, w = mul_Complex_Complex(w, wn)) {
				Complex x = a[j + k];
				Complex y = mul_Complex_Complex(w, a[j + k + i]);
				a[j + k] = add_Complex_Complex(x, y);
				a[j + k + i] = sub_Complex_Complex(x, y);
			}
		}
	}

	for(i = 0; i < len; i++) {
		if(opt == 1) {
			c[i] = a[i];
		} else if(opt == -1) {
			Complex t = mul_Complex_Complex(a[i], (Complex){cos(i * PI / len), -sin(i * PI / len)});
			z[i] = t.a / len;
		}
	}
}
