#include "stdio.h"
#include "sig_api.h"

extern long long n, k;

mat_poly_R *_a = NULL;

int sig_keygen(unsigned char *pk, unsigned char *sk) {
	mat_poly_R *_e = NULL, *_r = NULL;

	KeyGen(&_a, &_e, &_r);
	
	long long i;
	for(i = 0; i < (k + 2); i++) {
		long long j;
		for(j = 0; j < n; j++) {
			long long l = *((_a -> coefficient + i) -> coefficient + j);
			
			if(l < 0) {
				pk[i * n * 4 + j * 4 + 3] = 128;
				l = -l;
			} else {
				pk[i * n * 4 + j * 4 + 3] = 0;
			}
			
			pk[i * n * 4 + j * 4] = l % 256;
			pk[i * n * 4 + j * 4 + 1] = (l >> 8) % 256;
			pk[i * n * 4 + j * 4 + 2] = (l >> 16) % 256;
			pk[i * n * 4 + j * 4 + 3] += (l >> 24) % 256;
		}
	}
	
	for(i = 0; i < k; i++) {
		long long j;
		for(j = 0; j < n; j++) {
			long long l = *((_e -> coefficient + i) -> coefficient + j);
			
			if(l < 0) {
				sk[i * n * 4 + j * 4 + 3] = 128;
				l = -l;
			} else {
				sk[i * n * 4 + j * 4 + 3] = 0;
			}
			
			sk[i * n * 4 + j * 4] = l % 256;
			sk[i * n * 4 + j * 4 + 1] = (l >> 8) % 256;
			sk[i * n * 4 + j * 4 + 2] = (l >> 16) % 256;
			sk[i * n * 4 + j * 4 + 3] += (l >> 24) % 256;
		}
	}
	
	long long start = n * k * 4 / 2;
	for(i = 0; i < k; i++) {
		long long j;
		for(j = 0; j < n; j++) {
			long long l = *((_r -> coefficient + i) -> coefficient + j);
			
			if(l < 0) {
				sk[start + i * n * 4 + j * 4 + 3] = 128;
				l = -l;
			} else {
				sk[start + i * n * 4 + j * 4 + 3] = 0;
			}
			
			sk[start + i * n * 4 + j * 4] = l % 256;
			sk[start + i * n * 4 + j * 4 + 1] = (l >> 8) % 256;
			sk[start + i * n * 4 + j * 4 + 2] = (l >> 16) % 256;
			sk[start + i * n * 4 + j * 4 + 3] += (l >> 24) % 256;
		}
	}
	
	return 0;
}

int sig_sign(unsigned char *sk, unsigned char *m, unsigned long long mlen, unsigned char *sm, unsigned long long *smlen) {
	mat_poly_R _e = create_mat_poly_R(n, 1, k);
	mat_poly_R _r = create_mat_poly_R(n, 1, k);
	
	long long i;
	for(i = 0; i < k; i++) {
		long long j;
		for(j = 0; j < n; j++) {
			if(sk[i * n * 4 + j * 4 + 3] >= 128) {
				*((_e.coefficient + i) -> coefficient + j) = sk[i * n * 4 + j * 4 + 3] - 128;
			} else {
				*((_e.coefficient + i) -> coefficient + j) = sk[i * n * 4 + j * 4 + 3];
			}
			
			*((_e.coefficient + i) -> coefficient + j) <<= 8;
			*((_e.coefficient + i) -> coefficient + j) += sk[i * n * 4 + j * 4 + 2];
			*((_e.coefficient + i) -> coefficient + j) <<= 8;
			*((_e.coefficient + i) -> coefficient + j) += sk[i * n * 4 + j * 4 + 1];
			*((_e.coefficient + i) -> coefficient + j) <<= 8;
			*((_e.coefficient + i) -> coefficient + j) += sk[i * n * 4 + j * 4];
			
			if(sk[i * n * 4 + j * 4 + 3] >= 128) {
				*((_e.coefficient + i) -> coefficient + j) = - *((_e.coefficient + i) -> coefficient + j);
			}
		}
	}
	
	long long start = n * k * 4 / 2;
	for(i = 0; i < k; i++) {
		long long j;
		for(j = 0; j < n; j++) {
			if(sk[start + i * n * 4 + j * 4 + 3] >= 128) {
				*((_r.coefficient + i) -> coefficient + j) = sk[start + i * n * 4 + j * 4 + 3] - 128;
			} else {
				*((_r.coefficient + i) -> coefficient + j) = sk[start + i * n * 4 + j * 4 + 3];
			}
			
			*((_r.coefficient + i) -> coefficient + j) <<= 8;
			*((_r.coefficient + i) -> coefficient + j) += sk[start + i * n * 4 + j * 4 + 2];
			*((_r.coefficient + i) -> coefficient + j) <<= 8;
			*((_r.coefficient + i) -> coefficient + j) += sk[start + i * n * 4 + j * 4 + 1];
			*((_r.coefficient + i) -> coefficient + j) <<= 8;
			*((_r.coefficient + i) -> coefficient + j) += sk[start + i * n * 4 + j * 4];
			
			if(sk[start + i * n * 4 + j * 4 + 3] >= 128) {
				*((_r.coefficient + i) -> coefficient + j) = - *((_r.coefficient + i) -> coefficient + j);
			}
		}
	}
	
	mat_poly_R sig = Sign(_a, &_e, &_r, m, mlen);
	
	*smlen = n * (k + 2) * 4;
	
	for(i = 0; i < (k + 2); i++) {
		long long j;
		for(j = 0; j < n; j++) {
			long long l = *((sig.coefficient + i) -> coefficient + j);

			if(l < 0) {
				sm[i * n * 4 + j * 4 + 3] = 128;
				l = -l;
			} else {
				sm[i * n * 4 + j * 4 + 3] = 0;
			}

			sm[i * n * 4 + j * 4] = l % 256;
			sm[i * n * 4 + j * 4 + 1] = (l >> 8) % 256;
			sm[i * n * 4 + j * 4 + 2] = (l >> 16) % 256;
			sm[i * n * 4 + j * 4 + 3] += (l >> 24) % 256;
		}
	}
	
	return 0;
}

int sig_verf(unsigned char *pk, unsigned char *sm, unsigned long long smlen, unsigned char *m, unsigned long long mlen) {
	mat_poly_R _a = create_mat_poly_R(n, 1, k + 2);
	mat_poly_R sig = create_mat_poly_R(n, 1, k + 2);
	
	long long i;
	for(i = 0; i < k + 2; i++) {
		long long j;
		for(j = 0; j < n; j++) {
			if(pk[i * n * 4 + j * 4 + 3] >= 128) {
				*((_a.coefficient + i) -> coefficient + j) = pk[i * n * 4 + j * 4 + 3] - 128;
			} else {
				*((_a.coefficient + i) -> coefficient + j) = pk[i * n * 4 + j * 4 + 3];
			}

			*((_a.coefficient + i) -> coefficient + j) <<= 8;
			*((_a.coefficient + i) -> coefficient + j) += pk[i * n * 4 + j * 4 + 2];
			*((_a.coefficient + i) -> coefficient + j) <<= 8;
			*((_a.coefficient + i) -> coefficient + j) += pk[i * n * 4 + j * 4 + 1];
			*((_a.coefficient + i) -> coefficient + j) <<= 8;
			*((_a.coefficient + i) -> coefficient + j) += pk[i * n * 4 + j * 4];
			
			if(pk[i * n * 4 + j * 4 + 3] >= 128) {
				*((_a.coefficient + i) -> coefficient + j) = - *((_a.coefficient + i) -> coefficient + j);
			}
		}
	}
	
	for(i = 0; i < k + 2; i++) {
		long long j;
		for(j = 0; j < n; j++) {
			if(sm[i * n * 4 + j * 4 + 3] >= 128) {
				*((sig.coefficient + i) -> coefficient + j) = sm[i * n * 4 + j * 4 + 3] - 128;
			} else {
				*((sig.coefficient + i) -> coefficient + j) = sm[i * n * 4 + j * 4 + 3];
			}

			*((sig.coefficient + i) -> coefficient + j) <<= 8;
			*((sig.coefficient + i) -> coefficient + j) += sm[i * n * 4 + j * 4 + 2];
			*((sig.coefficient + i) -> coefficient + j) <<= 8;
			*((sig.coefficient + i) -> coefficient + j) += sm[i * n * 4 + j * 4 + 1];
			*((sig.coefficient + i) -> coefficient + j) <<= 8;
			*((sig.coefficient + i) -> coefficient + j) += sm[i * n * 4 + j * 4];
			
			if(sm[i * n * 4 + j * 4 + 3] >= 128) {
				*((sig.coefficient + i) -> coefficient + j) = - *((sig.coefficient + i) -> coefficient + j);
			}
		}
	}
	
	return Verify(&_a, &sig, m, mlen);
}
