#include <stdio.h>
#include <stdlib.h>
#include "util.h"
#include "fft.h"


long long mod_add_R(const long long a, const long long b) {
	extern long long modulo;
	long long result = (a + b) % modulo;
	return result >= 0 ? (result > modulo / 2 ? result - modulo : result) : (result < -modulo / 2 ? result + modulo : result);
}

long long mod_sub_R(const long long a, const long long b) {
	extern long long modulo;
	long long result = (a - b) % modulo;
	return result >= 0 ? (result > modulo / 2 ? result - modulo : result) : (result < -modulo / 2 ? result + modulo : result);
}

/*long long mod_mul_R(const long long a, const long long b) {
	long long result = (a * b) % modulo;
	return result >= 0 ? (result > modulo / 2 ? result - modulo : result) : (result < -modulo / 2 ? result + modulo : result);
}*/

//poly & mat_poly
//Initialization
poly_R create_poly_R(long long degree) {
	poly_R result;

	result.degree = degree;
	result.coefficient = (long long*)malloc(sizeof(long long) * degree);

	long long i;
	for(i = 0; i < degree; i++) {
		*(result.coefficient + i) = 0;
	}

	return result;
}

poly_K create_poly_K(long long degree) {
	poly_K result;

	result.degree = degree;
	result.coefficient = (double*)malloc(sizeof(double) * degree);

	long long i;
	for(i = 0; i < degree; i++) {
		*(result.coefficient + i) = 0;
	}

	return result;
}

mat_poly_R create_mat_poly_R(long long degree, long long row_size, long long column_size) {
	mat_poly_R result;

	result.row_size = row_size;
	result.column_size = column_size;
	result.coefficient = (poly_R*)malloc(sizeof(poly_R) * row_size * column_size);

	long long i;
	for(i = 0; i < result.row_size; i++) {
		long long j;
		for(j = 0; j < result.column_size; j++) {
			*(result.coefficient + i * result.column_size + j) = create_poly_R(degree);
		}
	}

	return result;
}

mat_poly_K create_mat_poly_K(long long degree, long long row_size, long long column_size) {
	mat_poly_K result;

	result.row_size = row_size;
	result.column_size = column_size;
	result.coefficient = (poly_K*)malloc(sizeof(poly_K) * row_size * column_size);

	long long i;
	for(i = 0; i < result.row_size; i++) {
		long long j;
		for(j = 0; j < result.column_size; j++) {
			*(result.coefficient + i * result.column_size + j) = create_poly_K(degree);
		}
	}

	return result;
}

/*poly_R create_random_poly_R(long long degree) {
	poly_R result;

	result.degree = degree;
	result.coefficient = (long long*)malloc(sizeof(long long) * degree);

	long long i;
	for(i = 0; i < degree; i++) {
		*(result.coefficient + i) = rand() % modulo;
	}

	return result;
}

poly_K create_random_poly_K(long long degree) {
	poly_K result;

	result.degree = degree;
	result.coefficient = (double*)malloc(sizeof(double) * degree);

	long long i;
	for(i = 0; i < degree; i++) {
		*(result.coefficient + i) = rand() % modulo;
	}

	return result;
}

mat_poly_R create_random_mat_poly_R(long long degree, long long row_size, long long column_size) {
	mat_poly_R result;

	result.row_size = row_size;
	result.column_size = column_size;
	result.coefficient = (poly_R*)malloc(sizeof(poly_R) * row_size * column_size);

	long long i;
	for(i = 0; i < result.row_size; i++) {
		long long j;
		for(j = 0; j < result.column_size; j++) {
			*(result.coefficient + i * result.column_size + j) = create_random_poly_R(degree);
		}
	}

	return result;
}

mat_poly_K create_random_mat_poly_K(long long degree, long long row_size, long long column_size) {
	mat_poly_K result;

	result.row_size = row_size;
	result.column_size = column_size;
	result.coefficient = (poly_K*)malloc(sizeof(poly_K) * row_size * column_size);

	long long i;
	for(i = 0; i < result.row_size; i++) {
		long long j;
		for(j = 0; j < result.column_size; j++) {
			*(result.coefficient + i * result.column_size + j) = create_random_poly_K(degree);
		}
	}

	return result;
}*/

//Output
void print_poly_R(poly_R a) {
	printf("[");
	
	long long i;
	for(i = 0; i < a.degree; i++) {
		printf("%I64d ", *(a.coefficient + i));
	}
	
	printf("]");
}

void print_poly_K(poly_K a) {
	printf("[");

	long long i;
	for(i = 0; i < a.degree; i++) {
		printf("%f ", *(a.coefficient + i));
	}

	printf("]");
}

void print_mat_poly_R(mat_poly_R a) {
	printf("[");
	
	long long degree = a.coefficient -> degree;
	
	long long i;
	for(i = 0; i < a.row_size; i++) {
		long long j;
		for(j = 0; j < degree; j++) {
			long long k;
			for(k = 0; k < a.column_size; k++) {
				printf("%d ", *((a.coefficient + i * a.column_size + k) -> coefficient + j));
			}
			printf("\n");
		}
	}
	
	printf("]");
}

void print_mat_poly_K(mat_poly_K a) {
	printf("[");

	long long degree = a.coefficient -> degree;

	long long i;
	for(i = 0; i < a.row_size; i++) {
		long long j;
		for(j = 0; j < degree; j++) {
			long long k;
			for(k = 0; k < a.column_size; k++) {
				printf("%f ", *((a.coefficient + i * a.column_size + k) -> coefficient + j));
			}
			printf("\n");
		}
	}

	printf("]");
}

int compare_poly_K(const poly_K a, const poly_K b) {
	if(a.degree != b.degree) {
		return a.degree > b.degree ? 1 : -1;
	} else {
		long long i;
		for(i = a.degree - 1; i >= 0; i--) {
			if(*(a.coefficient + i) == 0 && *(b.coefficient + i) != 0) {
				return -1;
			} else if(*(b.coefficient + i) == 0 && *(a.coefficient + i) != 0) {
				return 1;
			} else if(*(a.coefficient + i) != 0 && *(b.coefficient + i) != 0) {
				return 0;
			}
		}
		
		return 0;
	}
}

poly_K transpose_poly_K(poly_K a) {
	poly_K result = (poly_K){a.degree, NULL};
	result.coefficient = (double*)malloc(sizeof(double) * result.degree);
	
	*result.coefficient = *a.coefficient;
	
	long long i;
	for(i = 1; i < result.degree; i++) {
		*(result.coefficient + i) = - *(a.coefficient + a.degree - i);
	}
	
	return result;
}

poly_R mod_add_poly_R(const poly_R a, const poly_R b) {
	/*poly_R result = (poly_R){a.degree, NULL};
	result.coefficient = (long long*)malloc(sizeof(long long) * a.degree);*/
	
	poly_R result = create_poly_R(a.degree);
	
	long long i;
	for(i = 0; i < result.degree; i++) {
		*(result.coefficient + i) = mod_add_R(*(a.coefficient + i), *(b.coefficient + i));
	}
	
	return result;
}
poly_R mod_sub_poly_R(const poly_R a, const poly_R b) {
	/*poly_R result = (poly_R){a.degree, NULL};
	result.coefficient = (long long*)malloc(sizeof(long long) * a.degree);*/

	poly_R result = create_poly_R(a.degree);

	long long i;
	for(i = 0; i < result.degree; i++) {
		*(result.coefficient + i) = mod_sub_R(*(a.coefficient + i), *(b.coefficient + i));
	}

	return result;
}

poly_R mod_mul_poly_R(const poly_R p1, const poly_R p2) {
	extern long long modulo;
	
	poly_R result = (poly_R){p1.degree, NULL};
	result.coefficient = (long long*)malloc(sizeof(long long) * p1.degree);
	
	long long *a = (long long*)malloc(sizeof(long long) * p1.degree);
	long long *b = (long long*)malloc(sizeof(long long) * p2.degree);
	
	long long *product = (long long*)malloc(sizeof(long long) * result.degree);
	
	FFT_R(p1.coefficient, a, p1.degree, 1);
	FFT_R(p2.coefficient, b, p2.degree, 1);
	
	long long i;
	for(i = 0; i < p1.degree; i++) {
		product[i] = (a[i] * b[i]) % modulo;
	}
	
	FFT_R(product, result.coefficient, result.degree, -1);
	/*poly_R result = (poly_R){a.degree, NULL};
	result.coefficient = (long long*)malloc(sizeof(long long) * a.degree);
	
	if(result.degree < (1 << 8)) {
		Complex *_a = (Complex*)malloc(sizeof(Complex) * a.degree);
		Complex *_b = (Complex*)malloc(sizeof(Complex) * b.degree);
		Complex *_result = (Complex*)malloc(sizeof(Complex) * a.degree);

		FFT_R(a.coefficient, _a, a.degree, 1);
		FFT_R(b.coefficient, _b, b.degree, 1);

		long long i;
		for(i = 0; i < a.degree; i++) {
			*(_result + i) = mul_Complex_Complex(*(_a + i), *(_b + i));
		}

		FFT_R(result.coefficient, _result, result.degree, -1);
		
		for(i = 0; i < a.degree; i++) {
			*(result.coefficient + i) = *(result.coefficient + i) % modulo >= 0 ? (*(result.coefficient + i) % modulo > modulo / 2 ? *(result.coefficient + i) % modulo - modulo : *(result.coefficient + i) % modulo)
				: (*(result.coefficient + i) % modulo < -modulo / 2 ? *(result.coefficient + i) % modulo + modulo : *(result.coefficient + i) % modulo);
		}
	} else {
		double *_a1 = (double*)malloc(sizeof(double) * a.degree);
		double *_b1 = (double*)malloc(sizeof(double) * b.degree);
		double *_c1 = (double*)malloc(sizeof(double) * a.degree);
		
		Complex *_a2 = (Complex*)malloc(sizeof(Complex) * a.degree);
		Complex *_b2 = (Complex*)malloc(sizeof(Complex) * b.degree);
		Complex *_c2 = (Complex*)malloc(sizeof(Complex) * a.degree);
		
		long long i;
		for(i = 0; i < a.degree; i++) {
			*(_a1 + i) = *(a.coefficient + i) / (double)(1 << 9);
			*(_b1 + i) = *(b.coefficient + i) / (double)(1 << 9);
		}
		
		FFT_K(_a1, _a2, a.degree, 1);
		FFT_K(_b1, _b2, b.degree, 1);
		
		for(i = 0; i < a.degree; i++) {
			*(_c2 + i) = mul_Complex_Complex(*(_a2 + i), *(_b2 + i));
		}
		
		FFT_K(_c1, _c2, result.degree, -1);
		
		for(i = 0; i < a.degree; i++) {
			*(_c1 + i) -= (long long)(*(_c1 + i)) / modulo * modulo;
		}
		
		for(i = 0; i < a.degree; i++) {
			
			*(_c1 + i) *= (double)((long long)1 << 18);
			
			*(result.coefficient + i) = *(_c1 + i) >= 0 ? *(_c1 + i) + 0.5 : *(_c1 + i) - 0.5;
			*(result.coefficient + i) = *(result.coefficient + i) >= 0 ? (*(result.coefficient + i) % modulo > modulo / 2 ? *(result.coefficient + i) % modulo - modulo : *(result.coefficient + i) % modulo)
				: (*(result.coefficient + i) % modulo < -modulo / 2 ? *(result.coefficient + i) % modulo + modulo : *(result.coefficient + i) % modulo);
		}
	}
	
	return result;*/
	
	/*poly_R result = create_poly_R(p1.degree);

	long long i;
	for(i = 0; i < p1.degree; i++) {
		long long j;
		for(j = 0; j < p1.degree; j++) {
			if(i + j >= result.degree) {
				*(result.coefficient + i + j - result.degree) -= *(p1.coefficient + i) * *(p2.coefficient + j);
				*(result.coefficient + i + j - result.degree) = *(result.coefficient + i + j - result.degree) >= 0 ? (*(result.coefficient + i + j - result.degree) % modulo > modulo / 2 ? *(result.coefficient + i + j - result.degree) % modulo - modulo : *(result.coefficient + i + j - result.degree) % modulo)
					:(*(result.coefficient + i + j - result.degree) % modulo < -modulo / 2 ? *(result.coefficient + i + j - result.degree) % modulo + modulo : *(result.coefficient + i + j - result.degree) % modulo);
			} else {
				*(result.coefficient + i + j) += *(p1.coefficient + i) * *(p2.coefficient + j);
				*(result.coefficient + i + j) = *(result.coefficient + i + j) >= 0 ? (*(result.coefficient + i + j) % modulo > modulo / 2 ? *(result.coefficient + i + j) % modulo - modulo : *(result.coefficient + i + j) % modulo)
					:(*(result.coefficient + i + j) % modulo < -modulo / 2 ? *(result.coefficient + i + j) % modulo + modulo : *(result.coefficient + i + j) % modulo);
			}
		}*/
		
		/*for(j = 0; j < p1.degree; j++) {
			if(i + j >= result.degree) {
				*(result.coefficient + i + j - result.degree) -= mod_mul_R(*(p1.coefficient + i), *(p2.coefficient + j));
				*(result.coefficient + i + j - result.degree) %= modulo;
			} else {
				*(result.coefficient + i + j) += mod_mul_R(*(p1.coefficient + i), *(p2.coefficient + j));
				*(result.coefficient + i + j) %= modulo;
			}
		}*/
	//}
	
	/*for(i = 0; i < result.degree; i++) {
		*(result.coefficient + i) = *(result.coefficient + i) >= 0 ? (*(result.coefficient + i) % modulo > modulo / 2 ? *(result.coefficient + i) % modulo - modulo : *(result.coefficient + i)) :
			(*(result.coefficient + i) % modulo < -modulo / 2 ? *(result.coefficient + i) % modulo + modulo : *(result.coefficient + i) % modulo);
	}*/

	return result;
}

poly_K add_poly_K(const poly_K a, const poly_K b) {
	poly_K result = (poly_K){a.degree, NULL};
	result.coefficient = (double*)malloc(sizeof(double) * a.degree);

	long long i;
	for(i = 0; i < result.degree; i++) {
		*(result.coefficient + i) = *(a.coefficient + i) + *(b.coefficient + i);
	}

	return result;
}

poly_K sub_poly_K(const poly_K a, const poly_K b) {
	poly_K result = (poly_K){a.degree, NULL};
	result.coefficient = (double*)malloc(sizeof(double) * a.degree);

	long long i;
	for(i = 0; i < result.degree; i++) {
		*(result.coefficient + i) = *(a.coefficient + i) - *(b.coefficient + i);
	}

	return result;
}

poly_K mul_poly_K(const poly_K p1, const poly_K p2) {
	poly_K result = (poly_K){p1.degree, NULL};
	result.coefficient = (double*)malloc(sizeof(double) * p1.degree);
	
	Complex *a = (Complex*)malloc(sizeof(Complex) * p1.degree);
	Complex *b = (Complex*)malloc(sizeof(Complex) * p2.degree);
	Complex *c = (Complex*)malloc(sizeof(Complex) * result.degree);
	
	FFT_K(p1.coefficient, a, p1.degree, 1);
	FFT_K(p2.coefficient, b, p2.degree, 1);
	
	long long i;
	for(i = 0; i < result.degree; i++) {
		c[i] = mul_Complex_Complex(a[i], b[i]);
	}
	
	FFT_K(result.coefficient, c, result.degree, -1);
	
	return result;
}

poly_K div_poly_K(const poly_K p1, const poly_K p2) {
	/*poly_K result = create_poly_K(a.degree);
	
	long long deg_a = 0, deg_b = 0;
	
	long long i;
	for(i = a.degree - 1; i >= 0; i--) {
		if(*(a.coefficient + i) != 0) {
			deg_a = i;
			break;
		}
	}
	
	if(deg_a == 0) {
		*r = create_poly_K(a.degree);
		return result;
	}
	
	for(i = b.degree - 1; i >= 0; i--) {
		if(*(b.coefficient + i) != 0) {
			deg_b = i;
			break;
		}
	}
	
	if(deg_b == 0) {
		*r = (poly_K){0, NULL};
		return (poly_K){0, NULL};
	}
	
	if(deg_a < deg_b) {
		r -> degree = a.degree;

		long long i;
		for(i = 0; i < r -> degree; i++) {
			*(r -> coefficient + i) = *(a.coefficient + i);
		}

		return result;
	}
	
	poly_K dividend = (poly_K){a.degree, NULL}, divisior = (poly_K){b.degree, NULL};
	dividend.coefficient = (double*)malloc(sizeof(double) * dividend.degree);
	divisior.coefficient = (double*)malloc(sizeof(double) * divisior.degree);
	
	for(i = 0; i < dividend.degree; i++) {
		*(dividend.coefficient + i) = *(a.coefficient + i);
		*(divisior.coefficient + i) = *(b.coefficient + i);
	}
	
	for(i = deg_a - deg_b; i >= 0; i--) {
		*(result.coefficient + i) = *(dividend.coefficient + deg_b + i) / *(divisior.coefficient + deg_b);
		
		*(dividend.coefficient + deg_b + i) = 0;
		long long j;
		for(j = deg_b + i - 1; j >= 0; j--) {
			*(dividend.coefficient + j) -= *(result.coefficient + i) * *(divisior.coefficient + j - i);
		}
	}
	
	r -> degree = a.degree;
	for(i = 0; i < r -> degree; i++) {
		*(r -> coefficient + i) = *(dividend.coefficient + i);
	}
	
	return result;*/
	poly_K result = (poly_K){p1.degree, NULL};
	result.coefficient = (double*)malloc(sizeof(double) * p1.degree);

	Complex *a = (Complex*)malloc(sizeof(Complex) * p1.degree);
	Complex *b = (Complex*)malloc(sizeof(Complex) * p2.degree);
	Complex *c = (Complex*)malloc(sizeof(Complex) * result.degree);

	FFT_K(p1.coefficient, a, p1.degree, 1);
	FFT_K(p2.coefficient, b, p2.degree, 1);

	long long i;
	for(i = 0; i < result.degree; i++) {
		c[i] = div_Complex_Complex(a[i], b[i]);
	}

	FFT_K(result.coefficient, c, result.degree, -1);

	return result;
}

poly_K inv_poly_K(const poly_K a) {
	/*long long d;
	for(d = a.degree - 1; d >= 0; d--) {
		if(*(a.coefficient + d) != 0) {
			break;
		}
	}
	
	if(d == 0) {
		poly_K result = create_poly_K(a.degree);
		*result.coefficient = 1 / *a.coefficient;
		
		return result;
	} else if(d == -1) {
		return (poly_K){0, NULL};
	}
	
	poly_K result = (poly_K){a.degree, NULL};
	result.coefficient = (double*)malloc(sizeof(double) * result.degree);
	
	poly_K q[1025], r[1025], s[1025], t[1025];
	
	r[1] = a;
	s[0] = create_poly_K(a.degree);
	*s[0].coefficient = 1;
	s[1] = create_poly_K(a.degree);
	t[0] = create_poly_K(a.degree);
	t[1] = create_poly_K(a.degree);
	*t[1].coefficient = 1;
	
	r[2] = create_poly_K(a.degree);
	
	long long i;
	for(i = 0; i < d; i++) {
		*(r[2].coefficient + a.degree - 1 - i) = -1 / *(a.coefficient + d) * *(a.coefficient + d - 1 - i);
	}
	*r[2].coefficient += 1;
	
	q[2].coefficient = (double*)malloc(sizeof(double) * a.degree);
	q[2] = div_poly_K(r[2], a, &r[2]);
	*(q[2].coefficient + a.degree - d) = 1 / *(a.coefficient + d);
	
	for(i = 2; i < 1025; i++) {
		long long j;
		for(j = a.degree - 1; j >= 0; j--) {
			if(*(r[i].coefficient + j) != 0) {
				break;
			}
		}
		
		if(j <= 0) {
			break;
		}
		
		r[i + 1].coefficient = (double*)malloc(sizeof(double) * a.degree);
		q[i + 1].coefficient = (double*)malloc(sizeof(double) * a.degree);
		q[i + 1] = div_poly_K(r[i - 1], r[i], &r[i + 1]);
	}
	
	long long k;
	for(k = 2; k <= i; k++) {
		s[k] = sub_poly_K(s[k - 2], mul_poly_K(q[k], s[k - 1]));
		t[k] = sub_poly_K(t[k - 2], mul_poly_K(q[k], t[k - 1]));
	}
	
	for(k = 0; k < result.degree; k++) {
		*(result.coefficient + k) = *(t[i].coefficient + k) / *r[i].coefficient;
	}
	
	return result;*/
	poly_K result = (poly_K){a.degree, NULL};
	result.coefficient = (double*)malloc(sizeof(double) * result.degree);
	
	Complex *_a = (Complex*)malloc(sizeof(Complex) * a.degree);
	
	FFT_K(a.coefficient, _a, a.degree, 1);
	
	long long i;
	for(i = 0; i < a.degree; i++) {
		_a[i] = div_Complex_Complex((Complex){1, 0}, _a[i]);
	}
	
	FFT_K(result.coefficient, _a, result.degree, -1);
	
	return result;
}

poly_R inner_product_mat_poly_R(const mat_poly_R a, const mat_poly_R b) {
	if(a.row_size != 1 || b.row_size != 1 || a.column_size != b.column_size) {
		return (poly_R){0, NULL};
	} else {
		poly_R result = create_poly_R(a.coefficient -> degree);
		
		long long i;
		for(i = 0; i < a.column_size; i++) {
			poly_R product = mod_mul_poly_R(*(a.coefficient + i), *(b.coefficient + i));
			
			long long j;
			for(j = 0; j < result.degree; j++){
				*(result.coefficient + j) = mod_add_R(*(result.coefficient + j), *(product.coefficient + j));
			}
			//result = mod_add_poly_R(result, mod_mul_poly_R(*(a.coefficient + i), *(b.coefficient + i)));
		}
		
		return result;
	}
}

/*poly_R test(poly_R p1, poly_R p2) {
	poly_R result = create_poly_R(p1.degree);

	long long i;
	for(i = 0; i < p1.degree; i++) {
		long long j;
		for(j = 0; j < p1.degree; j++) {
			if(i + j >= result.degree) {
				*(result.coefficient + i + j - result.degree) -= *(p1.coefficient + i) * *(p2.coefficient + j);
				*(result.coefficient + i + j - result.degree) = *(result.coefficient + i + j - result.degree) >= 0 ? (*(result.coefficient + i + j - result.degree) % modulo > modulo / 2 ? *(result.coefficient + i + j - result.degree) % modulo - modulo : *(result.coefficient + i + j - result.degree) % modulo)
					:(*(result.coefficient + i + j - result.degree) % modulo < -modulo / 2 ? *(result.coefficient + i + j - result.degree) % modulo + modulo : *(result.coefficient + i + j - result.degree) % modulo);
			} else {
				*(result.coefficient + i + j) += *(p1.coefficient + i) * *(p2.coefficient + j);
				*(result.coefficient + i + j) = *(result.coefficient + i + j) >= 0 ? (*(result.coefficient + i + j) % modulo > modulo / 2 ? *(result.coefficient + i + j) % modulo - modulo : *(result.coefficient + i + j) % modulo)
					:(*(result.coefficient + i + j) % modulo < -modulo / 2 ? *(result.coefficient + i + j) % modulo + modulo : *(result.coefficient + i + j) % modulo);
			}
		}
	}

	return result;
}*/
