#ifndef _UTIL_HEADER
#define _UTIL_HEADER

typedef struct poly_R {
	//Element of ring R_q
	long long degree;
	long long *coefficient;
} poly_R;

typedef struct poly_K {
	//Element of field K
	long long degree;
	double *coefficient;
} poly_K;

typedef struct mat_poly_R {
	//matrix on Z or vector on R_q
	long long row_size;
	long long column_size;
	poly_R *coefficient;
} mat_poly_R;

typedef struct mat_poly_K{
	//matrix on Q or vector on K
	long long row_size;
	long long column_size;
	poly_K *coefficient;
} mat_poly_K;

long long mod_add_R(const long long a, const long long b);
long long mod_sub_R(const long long a, const long long b);
long long mod_mul_R(const long long a, const long long b);

//poly & mat
//Initialization
poly_R create_poly_R(long long degree);
poly_K create_poly_K(long long degree);
mat_poly_R create_mat_poly_R(long long degree, long long row_size, long long column_size);
mat_poly_K create_mat_poly_K(long long degree, long long row_size, long long column_size);

poly_R create_random_poly_R(long long degree);
poly_K create_random_poly_K(long long degree);
mat_poly_R create_random_mat_poly_R(long long degree, long long row_size, long long column_size);
mat_poly_K create_random_mat_poly_K(long long degree, long long row_size, long long column_size);

//Output
void print_poly_R(poly_R a);
void print_poly_K(poly_K a);
void print_mat_poly_R(mat_poly_R a);
void print_mat_poly_K(mat_poly_K a);

//Computation
int compare_poly_K(const poly_K a, const poly_K b);
poly_K transpose_poly_K(poly_K a);

poly_R mod_add_poly_R(const poly_R a, const poly_R b);
poly_R mod_sub_poly_R(const poly_R a, const poly_R b);
poly_R mod_mul_poly_R(const poly_R a, const poly_R b);

poly_K add_poly_K(const poly_K a, const poly_K b);
poly_K sub_poly_K(const poly_K a, const poly_K b);
poly_K mul_poly_K(const poly_K a, const poly_K b);

poly_K div_poly_K(const poly_K a, const poly_K b);
poly_K inv_poly_K(const poly_K a);

poly_R inner_product_mat_poly_R(const mat_poly_R a, const mat_poly_R b);

#endif
